"""Known decoding."""

import importlib


CACHE = {}
"""dict: a cache of known decoding"""


def get_decode(name):
    """
    Get decoding function by name.

    Decoding should be a function that takes
    a `bytes` input and returns a `str` result.

    :param str name: a decode name

    :return: the resulting decode
    :rtype: callable
    """

    decode = CACHE.get(name)

    if not decode:
        decode = CACHE[name] = importlib.import_module(".{0}".format(name), __name__).decode

    return decode



from unittest.mock import patch
import sys
import io

from content_hash.__main__ import main


CODEC = 'zeronet'
DECODED = '1HeLLo4uzjaLetFx6NH3PMwFP3qbRbTf3D'
ENCODED = 'e6013148654c4c6f34757a6a614c65744678364e4833504d774650337162526254663344'


def test_command_program_get_codec():
    argv = ['content-hash', 'get-codec', ENCODED]
    stdout = io.StringIO()

    with patch.object(sys, 'argv', argv), patch('sys.stdout', stdout):
        main()

    assert stdout.getvalue().split('\n')[0] == CODEC


def test_command_program_decode():
    argv = ['content-hash', 'decode', ENCODED]
    stdout = io.StringIO()

    with patch.object(sys, 'argv', argv), patch('sys.stdout', stdout):
        main()

    assert stdout.getvalue().split('\n')[0] == DECODED


def test_command_program_encode():
    argv = ['content-hash', 'encode', CODEC, DECODED]
    stdout = io.StringIO()

    with patch.object(sys, 'argv', argv), patch('sys.stdout', stdout):
        main()

    assert stdout.getvalue().split('\n')[0] == ENCODED



"""Known encoding."""

import importlib


CACHE = {}
"""dict: a cache of known encoding"""


def get_encode(name):
    """
    Get encoding function by name.

    Encoding should be a function that takes
    a `str` input and returns a `bytes` result.

    :param str name: an encode name

    :return: the resulting encode
    :rtype: callable
    """

    encode = CACHE.get(name)

    if not encode:
        encode = CACHE[name] = importlib.import_module(".{0}".format(name), __name__).encode

    return encode



"""Decode module for HEX multi hash."""

from multiformats import multihash, CID


def decode(value):
    """
    Decode HEX multi hash.

    :param bytes value: an encoded content

    :return: the decoded content
    :rtype: str
    """
    return multihash.unwrap(CID.decode(value).digest).hex()



from multiformats import multicodec

def raw_cid_value(cid):
    """Return raw representation of the CID as seen in multiformats_cid in bytes"""
    return b''.join([cid.version.to_bytes(1, byteorder='big'), multicodec.wrap(cid.codec, cid.digest)])



"""Encode module for Swarm."""

from multiformats import multihash, CID
from content_hash.utils import raw_cid_value

def encode(value):
    """
    Encode Swarm.

    :param bytes value: a decoded content

    :return: the encoded content
    :rtype: bytes
    """
    mhash = multihash.wrap(bytes.fromhex(value), 'keccak-256')
    cid = CID(base='base58btc', codec='swarm-manifest', version=1, digest=mhash)
    return raw_cid_value(cid)



import pytest

from content_hash.profiles import Profile


def dumb_decode(value):
    return value


def dumb_encode(value):
    return value


def test_profile_with_string():
    profile = Profile(decode='utf8', encode='utf8')

    assert profile.decode.__name__ == 'decode'
    assert profile.encode.__name__ == 'encode'


def test_profile_with_callable():
    profile = Profile(decode=dumb_decode, encode=dumb_encode)

    assert profile.decode.__name__ == 'dumb_decode'
    assert profile.encode.__name__ == 'dumb_encode'


def test_profile_with_other_decode():
    with pytest.raises(TypeError):
        profile = Profile(decode=True, encode='utf8')


def test_profile_with_other_encode():
    with pytest.raises(TypeError):
        profile = Profile(decode='utf8', encode=True)



import pytest

from multiformats.multicodec.err import MulticodecKeyError
import content_hash


def test_not_get_nonexistent_codec():
    with pytest.raises(ValueError):
        content_hash.get_codec('0001')


def test_not_decode_nonexistent_codec():
    with pytest.raises(ValueError):
        content_hash.decode('0001')


def test_not_encode_nonexistent_codec():
    with pytest.raises(MulticodecKeyError):
        content_hash.encode('this-codec-does-not-exist', 'value')



"""Encode module for UTF8."""

def encode(value):
    """
    Encode UTF8.

    :param bytes value: a decoded content

    :return: the encoded content
    :rtype: str
    """
    return value.encode('utf-8')



"""Decode module for UTF8."""

def decode(value):
    """
    Decode UTF8.

    :param bytes value: an encoded content

    :return: the decoded content
    :rtype: str
    """

    return value.decode('utf-8')



"""Known decoding and encoding profiles."""

from ..decodes import get_decode
from ..encodes import get_encode


class Profile:
    """Decoding and encoding profile"""

    __slots__ = ['__encode', '__decode']

    def __init__(self, decode, encode):
        """
        Initialize the class.

        :param Union[str, callable] decode: a profile decode
        :param Union[str, callable] encode: a profile encode
        """

        if isinstance(decode, str):
            self.__decode = get_decode(decode)
        elif callable(decode):
            self.__decode = decode
        else:
            raise TypeError('Argument `decode` must be a string or a callable')

        if isinstance(encode, str):
            self.__encode = get_encode(encode)
        elif callable(decode):
            self.__encode = encode
        else:
            raise TypeError('Argument `encode` must be a string or a callable')

    @property
    def decode(self):
        """
        Get profile decode.

        :return: the profile decode
        :rtype: callable
        """

        return self.__decode

    @property
    def encode(self):
        """
        Get profile encode.

        :return: the profile encode
        :rtype: str
        """

        return self.__encode


def get_profile(name):
    """
    Get profile by name.

    :param str name: a profile name

    :return: the profile
    :rtype: Profile
    """

    return PROFILES.get(name, PROFILES['default'])


PROFILES = {
    'ipfs': Profile(decode='b58_multi_hash', encode='ipfs'),
    'ipns': Profile(decode='b58_multi_hash', encode='ipfs'),
    'swarm': Profile(decode='hex_multi_hash', encode='swarm'),
    'default': Profile(decode='utf8', encode='utf8'),
}
"""
dict: a dict of known profiles

`encode` should be chosen among the `encode` modules
`decode` should be chosen among the `decode` modules
"""



"""Encode module for IPFS."""

import base58check
from multiformats import CID, multicodec
from multiformats_cid import make_cid
from content_hash.utils import raw_cid_value

def encode(value):
    """
    Encode IPFS.

    :param bytes value: a decoded content

    :return: the encoded content
    :rtype: bytes
    """
    mhash = base58check.b58decode(value)
    cid = CID(base='base58btc', codec='dag-pb', version=1, digest=mhash)
    return raw_cid_value(cid)



"""Command line program for EIP 1577 content hash."""

# pylint: disable=W0703

import argparse
import sys

from . import get_codec, decode, encode


def handle_get_codec(args):
    """Handle get codec."""

    try:
        chash = args.chash
        codec = get_codec(chash)

    except BaseException as err: # pragma: no cover
        print(str(err), file=sys.stderr)
        sys.exit(1)

    print(codec)


def handle_decode(args):
    """Handle decode."""

    try:
        chash = args.chash
        value = decode(chash)

    except BaseException as err: # pragma: no cover
        print(str(err), file=sys.stderr)
        sys.exit(1)

    print(value)


def handle_encode(args):
    """Handle encode."""

    try:
        codec = args.codec
        value = args.value
        chash = encode(codec, value)

    except BaseException as err: # pragma: no cover
        print(str(err), file=sys.stderr)
        sys.exit(1)

    print(chash)


def main():
    """Handle command line program."""

    parser = argparse.ArgumentParser(
        prog=__package__,
        description='Python implementation of EIP 1577 content hash'
    )
    subparsers = parser.add_subparsers()

    get_codec_parser = subparsers.add_parser(
        name='get-codec',
        help='extract the codec of a content hash'
    )
    get_codec_parser.set_defaults(which='get-codec')
    get_codec_parser.add_argument('chash', help='an hex string containing a content hash')

    get_codec_parser = subparsers.add_parser(
        name='decode',
        help='decode a content hash'
    )
    get_codec_parser.set_defaults(which='decode')
    get_codec_parser.add_argument('chash', help='an hex string containing a content hash')

    get_codec_parser = subparsers.add_parser(
        name='encode',
        help='encode a content hash'
    )
    get_codec_parser.set_defaults(which='encode')
    get_codec_parser.add_argument('codec', help='a codec of a content hash')
    get_codec_parser.add_argument('value', help='a value of a content hash')

    if len(sys.argv) == 1: # pragma: no cover
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    if args.which == 'get-codec':
        handle_get_codec(args)
    elif args.which == 'decode':
        handle_decode(args)
    elif args.which == 'encode':
        handle_encode(args)


if __name__ == '__main__': # pragma: no cover
    main()



"""Example for EIP 1577 content hash."""

import content_hash


def main():
    """Handle example."""

    # Get codec of content hash
    chash = 'e6013148654c4c6f34757a6a614c65744678364e4833504d774650337162526254663344'
    codec = content_hash.get_codec(chash)
    print('Codec of content hash:', codec)


    # Decode content hash
    chash = 'e6013148654c4c6f34757a6a614c65744678364e4833504d774650337162526254663344'
    address = content_hash.decode(chash)
    print('Decoded content hash:', address)

    # Encode content hash
    address = '1HeLLo4uzjaLetFx6NH3PMwFP3qbRbTf3D'
    chash = content_hash.encode('zeronet', address)
    print('Encoded content hash:', chash)


if __name__ == '__main__':
    main()



"""Decode module for B58 multi hash."""

from multiformats import multihash, CID
from base58check import b58encode

def decode(value):
    """
    Decode B58 multi hash.

    :param bytes value: an encoded content

    :return: the decoded content
    :rtype: str
    """
    return b58encode(CID.decode(value).digest)



from setuptools import setup


def readme():
    with open('README.md') as f:
            return f.read()

setup(
    name = 'rotki-content-hash',
    description = 'Python implementation of EIP 1577 content hash',
    long_description = readme(),
    long_description_content_type='text/markdown',
    license = 'MIT',
    version='0.0.1',
    setup_requires = ['setuptools-git-version'],

    packages = ['content_hash'],

    entry_points = {
        'console_scripts': ['rotki-content-hash=content_hash.__main__:main'],
    },

    install_requires = [
        'multiformats',
        'base58check',  # for b58 encode/decode
        'pysha3',  # for keccak-256
        'typing_extensions==4.5.0',  # due to https://github.com/hashberg-io/multiformats/issues/10
    ],

    extras_require = {
        'lint': ['pylint'],
        'test': ['pytest', 'pytest-cov'],
    },

    python_requires = '>= 3.5',

    author = 'Filip Š',
    author_email = 'projects@filips.si',
    url = 'https://github.com/filips123/ContentHashPy/',
    keywords = 'ethereum, ethereum-name-service, ens, eip1577, web3, decentralized',

    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: Name Service (DNS)',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
    ],
)



"""Python implementation of EIP 1577 content hash."""

from multiformats import multicodec

from .profiles import get_profile


def decode(chash):
    """
    Decode a content hash.

    :param str hash: a hex string containing a content hash

    :return: the decoded content
    :rtype: str
    """
    buffer = bytes.fromhex(chash.lstrip('0x'))

    codec, value = multicodec.unwrap(buffer)
    profile = get_profile(codec.name)
    result = profile.decode(value)
    if isinstance(result, bytes):
        result = result.decode()
    return result


def encode(codec, value):
    """
    Encode a content hash.

    :param str codec: a codec of a content hash
    :param str value: a value of a content hash

    :return: the resulting content hash
    :rtype: str
    """

    profile = get_profile(codec)

    value = profile.encode(value)
    value = multicodec.wrap(codec, value)
    return value.hex()


def get_codec(chash):
    """
    Extract the codec of a content hash

    :param str hash: a hex string containing a content hash

    :return: the extracted codec
    :rtype: str
    """

    buffer = bytes.fromhex(chash.lstrip('0x'))
    codec, _ =  multicodec.unwrap(buffer)
    return codec.name
